@@ -1,34 +1,56 @@
-from collections import Counter, Sized, Iterable
-from itertools import combinations, chain
+from collections import Counter
+from collections.abc import Iterable, Sized
+from itertools import chain, combinations
+from math import factorial, sqrt
 
-import numpy as np
-import math
 import scipy.spatial
-from math import factorial
+from numpy import abs as np_abs
+from numpy import (
+    array,
+    asarray,
+    average,
+    concatenate,
+    dot,
+    eye,
+    mean,
+    ones,
+    square,
+    subtract,
+)
+from numpy import sum as np_sum
+from numpy import zeros
+from numpy.linalg import det as ndet
+from numpy.linalg import matrix_rank, norm, slogdet, solve
 
 
 def fast_norm(v):
-    # notice this method can be even more optimised
-    if len(v) == 2:
-        return math.sqrt(v[0] * v[0] + v[1] * v[1])
-    if len(v) == 3:
-        return math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])
-    return math.sqrt(np.dot(v, v))
+    """Take the vector norm for len 2, 3 vectors.
+    Defaults to a square root of the dot product for larger vectors.
+
+    Note that for large vectors, it is possible for integer overflow to occur.
+    For instance:
+    vec = [49024, 59454, 12599, -63721, 18517, 27961]
+    dot(vec, vec) = -1602973744
+
+    """
+    len_v = len(v)
+    if len_v == 2:
+        return sqrt(v[0] * v[0] + v[1] * v[1])
+    if len_v == 3:
+        return sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])
+    return sqrt(dot(v, v))
 
 
 def fast_2d_point_in_simplex(point, simplex, eps=1e-8):
     (p0x, p0y), (p1x, p1y), (p2x, p2y) = simplex
     px, py = point
 
-    area = 0.5 * (- p1y * p2x + p0y * (p2x - p1x)
-                  + p1x * p2y + p0x * (p1y - p2y))
+    area = 0.5 * (-p1y * p2x + p0y * (p2x - p1x) + p1x * p2y + p0x * (p1y - p2y))
 
-    s = 1 / (2 * area) * (+ p0y * p2x + (p2y - p0y) * px
-                          - p0x * p2y + (p0x - p2x) * py)
+    s = 1 / (2 * area) * (+p0y * p2x + (p2y - p0y) * px - p0x * p2y + (p0x - p2x) * py)
     if s < -eps or s > 1 + eps:
         return False
-    t = 1 / (2 * area) * (+ p0x * p1y + (p0y - p1y) * px
-                          - p0y * p1x + (p1x - p0x) * py)
+    t = 1 / (2 * area) * (+p0x * p1y + (p0y - p1y) * px - p0y * p1x + (p1x - p0x) * py)
 
     return (t >= -eps) and (s + t <= 1 + eps)
 
@@ -37,9 +59,9 @@ def point_in_simplex(point, simplex, eps=1e-8):
     if len(point) == 2:
         return fast_2d_point_in_simplex(point, simplex, eps)
 
-    x0 = np.array(simplex[0], dtype=float)
-    vectors = np.array(simplex[1:], dtype=float) - x0
-    alpha = np.linalg.solve(vectors.T, point - x0)
+    x0 = array(simplex[0], dtype=float)
+    vectors = array(simplex[1:], dtype=float) - x0
+    alpha = solve(vectors.T, point - x0)
 
     return all(alpha > -eps) and sum(alpha) < 1 + eps
 
@@ -55,9 +77,9 @@ def fast_2d_circumcircle(points):
     Returns
     -------
     tuple
-        (center point : tuple(int), radius: int)
+        (center point : tuple(float), radius: float)
     """
-    points = np.array(points)
+    points = array(points)
     # transform to relative coordinates
     pts = points[1:] - points[0]
 
@@ -67,21 +89,21 @@ def fast_2d_circumcircle(points):
     l2 = x2 * x2 + y2 * y2
 
     # compute some determinants
-    dx = + l1 * y2 - l2 * y1
-    dy = - l1 * x2 + l2 * x1
-    aa = + x1 * y2 - x2 * y1
+    dx = +l1 * y2 - l2 * y1
+    dy = -l1 * x2 + l2 * x1
+    aa = +x1 * y2 - x2 * y1
     a = 2 * aa
 
     # compute center
     x = dx / a
     y = dy / a
-    radius = math.sqrt(x*x + y*y)  # radius = norm([x, y])
+    radius = sqrt(x * x + y * y)  # radius = norm([x, y])
 
     return (x + points[0][0], y + points[0][1]), radius
 
 
 def fast_3d_circumcircle(points):
-    """Compute the center and radius of the circumscribed shpere of a simplex.
+    """Compute the center and radius of the circumscribed sphere of a simplex.
 
     Parameters
     ----------
@@ -91,9 +113,9 @@ def fast_3d_circumcircle(points):
     Returns
     -------
     tuple
-        (center point : tuple(int), radius: int)
+        (center point : tuple(float), radius: float)
     """
-    points = np.array(points)
+    points = array(points)
     pts = points[1:] - points[0]
 
     (x1, y1, z1), (x2, y2, z2), (x3, y3, z3) = pts
@@ -103,41 +125,54 @@ def fast_3d_circumcircle(points):
     l3 = x3 * x3 + y3 * y3 + z3 * z3
 
     # Compute some determinants:
-    dx = (+ l1 * (y2 * z3 - z2 * y3)
-          - l2 * (y1 * z3 - z1 * y3)
-          + l3 * (y1 * z2 - z1 * y2))
-    dy = (+ l1 * (x2 * z3 - z2 * x3)
-          - l2 * (x1 * z3 - z1 * x3)
-          + l3 * (x1 * z2 - z1 * x2))
-    dz = (+ l1 * (x2 * y3 - y2 * x3)
-          - l2 * (x1 * y3 - y1 * x3)
-          + l3 * (x1 * y2 - y1 * x2))
-    aa = (+ x1 * (y2 * z3 - z2 * y3)
-          - x2 * (y1 * z3 - z1 * y3)
-          + x3 * (y1 * z2 - z1 * y2))
+    dx = +l1 * (y2 * z3 - z2 * y3) - l2 * (y1 * z3 - z1 * y3) + l3 * (y1 * z2 - z1 * y2)
+    dy = +l1 * (x2 * z3 - z2 * x3) - l2 * (x1 * z3 - z1 * x3) + l3 * (x1 * z2 - z1 * x2)
+    dz = +l1 * (x2 * y3 - y2 * x3) - l2 * (x1 * y3 - y1 * x3) + l3 * (x1 * y2 - y1 * x2)
+    aa = +x1 * (y2 * z3 - z2 * y3) - x2 * (y1 * z3 - z1 * y3) + x3 * (y1 * z2 - z1 * y2)
     a = 2 * aa
 
     center = (dx / a, -dy / a, dz / a)
     radius = fast_norm(center)
-    center = (center[0] + points[0][0],
-              center[1] + points[0][1],
-              center[2] + points[0][2])
+    center = (
+        center[0] + points[0][0],
+        center[1] + points[0][1],
+        center[2] + points[0][2],
+    )
 
     return center, radius
 
 
 def fast_det(matrix):
-    matrix = np.asarray(matrix, dtype=float)
+    matrix = asarray(matrix, dtype=float)
     if matrix.shape == (2, 2):
         return matrix[0][0] * matrix[1][1] - matrix[1][0] * matrix[0][1]
     elif matrix.shape == (3, 3):
         a, b, c, d, e, f, g, h, i = matrix.ravel()
-        return a * (e*i - f*h) - b * (d*i - f*g) + c * (d*h - e*g)
+        return a * (e * i - f * h) - b * (d * i - f * g) + c * (d * h - e * g)
     else:
-        return np.linalg.det(matrix)
+        return ndet(matrix)
 
 
 def circumsphere(pts):
+    """Compute the center and radius of a N dimension sphere which touches each point in pts.
+
+    Parameters
+    ----------
+    pts : array-like, of shape (N-dim + 1, N-dim)
+        The points for which we would like to compute a circumsphere.
+
+    Returns
+    -------
+    center : tuple of floats of size N-dim
+    radius : a positive float
+    A valid center and radius, if a circumsphere is possible, and no points are repeated.
+    If points are repeated, or a circumsphere is not possible, will return nans, and a
+    ZeroDivisionError may occur.
+    Will fail for matrices which are not (N-dim + 1, N-dim) in size due to non-square determinants:
+    will raise numpy.linalg.LinAlgError.
+    May fail for points that are integers (due to 32bit integer overflow).
+    """
+
     dim = len(pts) - 1
     if dim == 2:
         return fast_2d_circumcircle(pts)
@@ -145,20 +180,23 @@ def circumsphere(pts):
         return fast_3d_circumcircle(pts)
 
     # Modified method from http://mathworld.wolfram.com/Circumsphere.html
-    mat = [[np.sum(np.square(pt)), *pt, 1] for pt in pts]
-
-    center = []
+    mat = array([[np_sum(square(pt)), *pt, 1] for pt in pts])
+    center = zeros(dim)
+    a = 1 / (2 * ndet(mat[:, 1:]))
+    factor = a
+    # Use ind to index into the matrix columns
+    ind = ones((dim + 2,), bool)
     for i in range(1, len(pts)):
-        r = np.delete(mat, i, 1)
-        factor = (-1) ** (i + 1)
-        center.append(factor * fast_det(r))
-
-    a = fast_det(np.delete(mat, 0, 1))
-    center = [x / (2 * a) for x in center]
+        ind[i - 1] = True
+        ind[i] = False
+        center[i - 1] = factor * ndet(mat[:, ind])
+        factor *= -1
 
+    # Use subtract as we don't know the type of x0.
     x0 = pts[0]
-    vec = np.subtract(center, x0)
-    radius = fast_norm(vec)
+    vec = subtract(center, x0)
+    # Vector norm.
+    radius = sqrt(dot(vec, vec))
 
     return tuple(center), radius
 
@@ -182,8 +220,8 @@ def orientation(face, origin):
     If two points lie on the same side of the face, the orientation will
     be equal, if they lie on the other side of the face, it will be negated.
     """
-    vectors = np.array(face)
-    sign, logdet = np.linalg.slogdet(vectors - origin)
+    vectors = array(face)
+    sign, logdet = slogdet(vectors - origin)
     if logdet < -50:  # assume it to be zero when it's close to zero
         return 0
     return sign
@@ -206,7 +244,7 @@ def simplex_volume_in_embedding(vertices) -> float:
 
     Returns
     -------
-    volume : int
+    volume : float
         the volume of the simplex with given vertices.
 
     Raises
@@ -215,37 +253,36 @@ def simplex_volume_in_embedding(vertices) -> float:
         if the vertices do not form a simplex (for example,
         because they are coplanar, colinear or coincident).
     """
-
     # Implements http://mathworld.wolfram.com/Cayley-MengerDeterminant.html
     # Modified from https://codereview.stackexchange.com/questions/77593/calculating-the-volume-of-a-tetrahedron
 
-    vertices = np.asarray(vertices, dtype=float)
+    vertices = asarray(vertices, dtype=float)
     dim = len(vertices[0])
     if dim == 2:
         # Heron's formula
-        a, b, c = scipy.spatial.distance.pdist(vertices, metric='euclidean')
+        a, b, c = scipy.spatial.distance.pdist(vertices, metric="euclidean")
         s = 0.5 * (a + b + c)
-        return math.sqrt(s * (s - a) * (s - b) * (s - c))
+        return sqrt(s * (s - a) * (s - b) * (s - c))
 
     # β_ij = |v_i - v_k|²
-    sq_dists = scipy.spatial.distance.pdist(vertices, metric='sqeuclidean')
+    sq_dists = scipy.spatial.distance.pdist(vertices, metric="sqeuclidean")
 
     # Add border while compressed
     num_verts = scipy.spatial.distance.num_obs_y(sq_dists)
-    bordered = np.concatenate((np.ones(num_verts), sq_dists))
+    bordered = concatenate((ones(num_verts), sq_dists))
 
     # Make matrix and find volume
     sq_dists_mat = scipy.spatial.distance.squareform(bordered)
 
-    coeff = - (-2) ** (num_verts-1) * factorial(num_verts-1) ** 2
+    coeff = -((-2) ** (num_verts - 1)) * factorial(num_verts - 1) ** 2
     vol_square = fast_det(sq_dists_mat) / coeff
 
     if vol_square < 0:
-        if abs(vol_square) < 1e-15:
+        if vol_square > -1e-15:
             return 0
-        raise ValueError('Provided vertices do not form a simplex')
+        raise ValueError("Provided vertices do not form a simplex")
 
-    return np.sqrt(vol_square)
+    return sqrt(vol_square)
 
 
 class Triangulation:
@@ -296,10 +333,12 @@ class Triangulation:
             raise ValueError("Please provide at least one simplex")
 
         coords = list(map(tuple, coords))
-        vectors = np.subtract(coords[1:], coords[0])
-        if np.linalg.matrix_rank(vectors) < dim:
-            raise ValueError("Initial simplex has zero volumes "
-                             "(the points are linearly dependent)")
+        vectors = subtract(coords[1:], coords[0])
+        if matrix_rank(vectors) < dim:
+            raise ValueError(
+                "Initial simplex has zero volumes "
+                "(the points are linearly dependent)"
+            )
 
         self.vertices = list(coords)
         self.simplices = set()
@@ -325,7 +364,12 @@ class Triangulation:
             self.vertex_to_simplices[vertex].add(simplex)
 
     def get_vertices(self, indices):
-        return [self.vertices[i] for i in indices]
+        return [self.get_vertex(i) for i in indices]
+
+    def get_vertex(self, index):
+        if index is None:
+            return None
+        return self.vertices[index]
 
     def get_reduced_simplex(self, point, simplex, eps=1e-8) -> list:
         """Check whether vertex lies within a simplex.
@@ -340,9 +384,9 @@ class Triangulation:
         if len(simplex) != self.dim + 1:
             # We are checking whether point belongs to a face.
             simplex = self.containing(simplex).pop()
-        x0 = np.array(self.vertices[simplex[0]])
-        vectors = np.array(self.get_vertices(simplex[1:])) - x0
-        alpha = np.linalg.solve(vectors.T, point - x0)
+        x0 = array(self.vertices[simplex[0]])
+        vectors = array(self.get_vertices(simplex[1:])) - x0
+        alpha = solve(vectors.T, point - x0)
         if any(alpha < -eps) or sum(alpha) > 1 + eps:
             return []
 
@@ -385,8 +429,7 @@ class Triangulation:
         elif simplices is None:
             simplices = self.simplices
 
-        faces = (face for tri in simplices
-                 for face in combinations(tri, dim))
+        faces = (face for tri in simplices for face in combinations(tri, dim))
 
         if vertices is not None:
             return (face for face in faces if all(i in vertices for i in face))
@@ -406,7 +449,7 @@ class Triangulation:
         # we do not really need the center, we only need a point that is
         # guaranteed to lie strictly within the hull
         hull_points = self.get_vertices(self.hull)
-        pt_center = np.average(hull_points, axis=0)
+        pt_center = average(hull_points, axis=0)
 
         pt_index = len(self.vertices)
         self.vertices.append(new_vertex)
@@ -450,7 +493,7 @@ class Triangulation:
         tuple (center point, radius)
             The center and radius of the circumscribed circle
         """
-        pts = np.dot(self.get_vertices(simplex), transform)
+        pts = dot(self.get_vertices(simplex), transform)
         return circumsphere(pts)
 
     def point_in_cicumcircle(self, pt_index, simplex, transform):
@@ -458,13 +501,13 @@ class Triangulation:
         eps = 1e-8
 
         center, radius = self.circumscribed_circle(simplex, transform)
-        pt = np.dot(self.get_vertices([pt_index]), transform)[0]
+        pt = dot(self.get_vertices([pt_index]), transform)[0]
 
-        return np.linalg.norm(center - pt) < (radius * (1 + eps))
+        return norm(center - pt) < (radius * (1 + eps))
 
     @property
     def default_transform(self):
-        return np.eye(self.dim)
+        return eye(self.dim)
 
     def bowyer_watson(self, pt_index, containing_simplex=None, transform=None):
         """Modified Bowyer-Watson point adding algorithm.
@@ -506,17 +549,11 @@ class Triangulation:
                 bad_triangles.add(simplex)
 
                 # Get all simplices that share at least a point with the simplex
-                neighbours = set.union(*[self.vertex_to_simplices[p]
-                                         for p in todo_points])
+                neighbors = self.get_neighbors_from_vertices(todo_points)
                 # Filter out the already evaluated simplices
-                neighbours = neighbours - done_simplices
-
-                # Keep only the simplices sharing a whole face with the current simplex
-                neighbours = set(
-                    simpl for simpl in neighbours
-                    if len(set(simpl) & set(simplex)) == self.dim  # they share a face
-                )
-                queue.update(neighbours)
+                neighbors = neighbors - done_simplices
+                neighbors = self.get_face_sharing_neighbors(neighbors, simplex)
+                queue.update(neighbors)
 
         faces = list(self.faces(simplices=bad_triangles))
 
@@ -541,9 +578,9 @@ class Triangulation:
         volume is only dependent on the shape of the simplex and not on the
         absolute size. Due to the weird scaling, the only use of this method
         is to check that a simplex is almost flat."""
-        vertices = np.array(self.get_vertices(simplex))
+        vertices = array(self.get_vertices(simplex))
         vectors = vertices[1:] - vertices[0]
-        average_edge_length = np.mean(np.abs(vectors))
+        average_edge_length = mean(np_abs(vectors))
         return self.volume(simplex) / (average_edge_length ** self.dim)
 
     def add_point(self, point, simplex=None, transform=None):
@@ -572,8 +609,9 @@ class Triangulation:
             temporary_simplices = self._extend_hull(point)
 
             pt_index = len(self.vertices) - 1
-            deleted_simplices, added_simplices = \
-                self.bowyer_watson(pt_index, transform=transform)
+            deleted_simplices, added_simplices = self.bowyer_watson(
+                pt_index, transform=transform
+            )
 
             deleted = deleted_simplices - temporary_simplices
             added = added_simplices | (temporary_simplices - deleted_simplices)
@@ -582,7 +620,7 @@ class Triangulation:
             reduced_simplex = self.get_reduced_simplex(point, simplex)
             if not reduced_simplex:
                 self.vertex_to_simplices.pop()  # revert adding vertex
-                raise ValueError('Point lies outside of the specified simplex.')
+                raise ValueError("Point lies outside of the specified simplex.")
             else:
                 simplex = reduced_simplex
 
@@ -595,8 +633,8 @@ class Triangulation:
             return self.bowyer_watson(pt_index, actual_simplex, transform)
 
     def volume(self, simplex):
-        prefactor = np.math.factorial(self.dim)
-        vertices = np.array(self.get_vertices(simplex))
+        prefactor = factorial(self.dim)
+        vertices = array(self.get_vertices(simplex))
         vectors = vertices[1:] - vertices[0]
         return float(abs(fast_det(vectors)) / prefactor)
 
@@ -606,12 +644,10 @@ class Triangulation:
     def reference_invariant(self):
         """vertex_to_simplices and simplices are compatible."""
         for vertex in range(len(self.vertices)):
-            if any(vertex not in tri
-                   for tri in self.vertex_to_simplices[vertex]):
+            if any(vertex not in tri for tri in self.vertex_to_simplices[vertex]):
                 return False
         for simplex in self.simplices:
-            if any(simplex not in self.vertex_to_simplices[pt]
-                   for pt in simplex):
+            if any(simplex not in self.vertex_to_simplices[pt] for pt in simplex):
                 return False
         return True
 
@@ -619,6 +655,37 @@ class Triangulation:
         """Simplices originating from a vertex don't overlap."""
         raise NotImplementedError
 
+    def get_neighbors_from_vertices(self, simplex):
+        return set.union(*[self.vertex_to_simplices[p] for p in simplex])
+
+    def get_face_sharing_neighbors(self, neighbors, simplex):
+        """Keep only the simplices sharing a whole face with simplex."""
+        return {
+            simpl for simpl in neighbors if len(set(simpl) & set(simplex)) == self.dim
+        }  # they share a face
+
+    def get_simplices_attached_to_points(self, indices):
+        # Get all simplices that share at least a point with the simplex
+        neighbors = self.get_neighbors_from_vertices(indices)
+        return self.get_face_sharing_neighbors(neighbors, indices)
+
+    def get_opposing_vertices(self, simplex):
+        if simplex not in self.simplices:
+            raise ValueError("Provided simplex is not part of the triangulation")
+        neighbors = self.get_simplices_attached_to_points(simplex)
+
+        def find_opposing_vertex(vertex):
+            # find the simplex:
+            simp = next((x for x in neighbors if vertex not in x), None)
+            if simp is None:
+                return None
+            opposing = set(simp) - set(simplex)
+            assert len(opposing) == 1
+            return opposing.pop()
+
+        result = tuple(find_opposing_vertex(v) for v in simplex)
+        return result
+
     @property
     def hull(self):
         """Compute hull from triangulation.
@@ -636,11 +703,12 @@ class Triangulation:
         """
         counts = Counter(self.faces())
         if any(i > 2 for i in counts.values()):
-            raise RuntimeError("Broken triangulation, a (N-1)-dimensional"
-                               " appears in more than 2 simplices.")
+            raise RuntimeError(
+                "Broken triangulation, a (N-1)-dimensional"
+                " appears in more than 2 simplices."
+            )
 
-        hull = set(point for face, count in counts.items() if count == 1
-                   for point in face)
+        hull = {point for face, count in counts.items() if count == 1 for point in face}
         return hull
 
     def convex_invariant(self, vertex):
