@@ -1,16 +1,16 @@
 # Copyright 2010 Pedro Gonnet
 # Copyright 2017 Christoph Groth
 
-import warnings
-from fractions import Fraction as Frac
 from collections import defaultdict
+from fractions import Fraction as Frac
+
 import numpy as np
 from numpy.testing import assert_allclose
-from scipy.linalg import norm, inv
-
+from scipy.linalg import inv, norm
 
 eps = np.spacing(1)
 
+
 def legendre(n):
     """Return the first n Legendre polynomials.
 
@@ -25,7 +25,7 @@ def legendre(n):
     for i in range(2, n):
         # Use Bonnet's recursion formula.
         new = (i + 1) * [Frac(0)]
-        new[1:] = (r * (2*i - 1) for r in result[-1])
+        new[1:] = (r * (2 * i - 1) for r in result[-1])
         new[:-2] = (n - r * (i - 1) for n, r in zip(new[:-2], result[-2]))
         new[:] = (n / i for n in new)
         result.append(new)
@@ -50,7 +50,7 @@ def newton(n):
     # term m * cos(a * Pi / n) to be added to prefactor of the
     # monomial x^(n-d).
 
-    mod = 2 * (n-1)
+    mod = 2 * (n - 1)
     terms = defaultdict(int)
     terms[0, 0] += 1
 
@@ -61,7 +61,7 @@ def newton(n):
                 # In order to reduce the number of terms, cosine
                 # arguments are mapped back to the inteval [0, pi/2).
                 arg = (a + b) % mod
-                if arg > n-1:
+                if arg > n - 1:
                     arg = mod - arg
                 if arg >= n // 2:
                     if n % 2 and arg == n // 2:
@@ -83,9 +83,9 @@ def newton(n):
         # c[n - d] += m * np.cos(a * np.pi / (n - 1))
 
     cf = np.array(c, float)
-    assert all(int(cfe) == ce for cfe, ce in zip(cf, c)), 'Precision loss'
+    assert all(int(cfe) == ce for cfe, ce in zip(cf, c)), "Precision loss"
 
-    cf /= 2.**np.arange(n, -1, -1)
+    cf /= 2.0 ** np.arange(n, -1, -1)
     return cf
 
 
@@ -124,13 +124,13 @@ def calc_bdef(ns):
     for n in ns:
         poly = []
         a = list(map(Frac, newton(n)))
-        for b in legs[:n + 1]:
+        for b in legs[: n + 1]:
             igral = scalar_product(a, b)
 
             # Normalize & store.  (The polynomials returned by
             # legendre() have standard normalization that is not
             # orthonormal.)
-            poly.append(np.sqrt((2*len(b) - 1) / 2) * igral)
+            poly.append(np.sqrt((2 * len(b) - 1) / 2) * igral)
 
         result.append(np.array(poly))
     return result
@@ -138,7 +138,7 @@ def calc_bdef(ns):
 
 # Nodes and Newton polynomials.
 n = (5, 9, 17, 33)
-xi = [-np.cos(np.arange(n[j])/(n[j]-1) * np.pi) for j in range(4)]
+xi = [-np.cos(np.arange(n[j]) / (n[j] - 1) * np.pi) for j in range(4)]
 # Make `xi` perfectly anti-symmetric, important for splitting the intervals
 xi = [(row - row[::-1]) / 2 for row in xi]
 
@@ -148,11 +148,12 @@ b_def = calc_bdef(n)
 def calc_V(xi, n):
     V = [np.ones(xi.shape), xi.copy()]
     for i in range(2, n):
-        V.append((2*i-1) / i * xi * V[-1] - (i-1) / i * V[-2])
+        V.append((2 * i - 1) / i * xi * V[-1] - (i - 1) / i * V[-2])
     for i in range(n):
         V[i] *= np.sqrt(i + 0.5)
     return np.array(V).T
 
+
 # Compute the Vandermonde-like matrix and its inverse.
 V = [calc_V(*args) for args in zip(xi, n)]
 V_inv = list(map(inv, V))
@@ -178,8 +179,9 @@ sqrt_one_half = np.sqrt(0.5)
 
 
 k = np.arange(n[3])
-alpha = np.sqrt((k+1)**2 / (2*k+1) / (2*k+3))
-gamma = np.concatenate([[0, 0], np.sqrt(k[2:]**2 / (4*k[2:]**2-1))])
+alpha = np.sqrt((k + 1) ** 2 / (2 * k + 1) / (2 * k + 3))
+gamma = np.concatenate([[0, 0], np.sqrt(k[2:] ** 2 / (4 * k[2:] ** 2 - 1))])
+
 
 def _downdate(c, nans, depth):
     # This is algorithm 5 from the thesis of Pedro Gonnet.
@@ -190,8 +192,7 @@ def _downdate(c, nans, depth):
         xii = xi[depth][i]
         b[m] = (b[m] + xii * b[m + 1]) / alpha[m - 1]
         for j in range(m - 1, 0, -1):
-            b[j] = ((b[j] + xii * b[j + 1] - gamma[j + 1] * b[j + 2])
-                    / alpha[j - 1])
+            b[j] = (b[j] + xii * b[j + 1] - gamma[j + 1] * b[j + 2]) / alpha[j - 1]
         b = b[1:]
 
         c[:m] -= c[m] / b[m] * b[:m]
@@ -227,8 +228,7 @@ class DivergentIntegralError(ValueError):
 
 
 class _Interval:
-    __slots__ = ['a', 'b', 'c', 'fx', 'igral', 'err', 'depth', 'rdepth', 'ndiv',
-                 'c00']
+    __slots__ = ["a", "b", "c", "fx", "igral", "err", "depth", "rdepth", "ndiv", "c00"]
 
     def __init__(self, a, b, depth, rdepth):
         self.a = a
@@ -254,8 +254,8 @@ class _Interval:
     def calc_igral_and_err(self, c_old):
         self.c = c_new = _calc_coeffs(self.fx, self.depth)
         c_diff = np.zeros(max(len(c_old), len(c_new)))
-        c_diff[:len(c_old)] = c_old
-        c_diff[:len(c_new)] -= c_new
+        c_diff[: len(c_old)] = c_old
+        c_diff[: len(c_new)] -= c_new
         c_diff = norm(c_diff)
         w = self.b - self.a
         self.igral = w * c_new[0] * sqrt_one_half
@@ -267,8 +267,7 @@ class _Interval:
         f_center = self.fx[(len(self.fx) - 1) // 2]
 
         rdepth = self.rdepth + 1
-        ivals = [_Interval(self.a, m, 0, rdepth),
-                 _Interval(m, self.b, 0, rdepth)]
+        ivals = [_Interval(self.a, m, 0, rdepth), _Interval(m, self.b, 0, rdepth)]
         points = np.concatenate([ival.points()[1:-1] for ival in ivals])
         nr_points = len(points)
         fxs = np.empty((2, n[0]))
@@ -278,11 +277,11 @@ class _Interval:
 
         for ival, fx, T in zip(ivals, fxs, T_lr):
             ival.fx = fx
-            ival.calc_igral_and_err(T[:, :self.c.shape[0]] @ self.c)
+            ival.calc_igral_and_err(T[:, : self.c.shape[0]] @ self.c)
 
             ival.c00 = ival.c[0]
-            ival.ndiv = (self.ndiv + (self.c00 and ival.c00 / self.c00 > 2))
-            if ival.ndiv > ndiv_max and 2*ival.ndiv > ival.rdepth:
+            ival.ndiv = self.ndiv + (self.c00 and ival.c00 / self.c00 > 2)
+            if ival.ndiv > ndiv_max and 2 * ival.ndiv > ival.rdepth:
                 # Signal a divergent integral.
                 return (ival.a, ival.b, ival.b - ival.a), nr_points
 
@@ -293,14 +292,14 @@ class _Interval:
         self.depth = depth = self.depth + 1
         points = self.points()
         fx = np.empty(n[depth])
-        fx[0:n[depth]:2] = self.fx
-        fx[1:n[depth]-1:2] = f(points[1:n[depth]-1:2])
+        fx[0 : n[depth] : 2] = self.fx
+        fx[1 : n[depth] - 1 : 2] = f(points[1 : n[depth] - 1 : 2])
         self.fx = fx
         split = self.calc_igral_and_err(self.c) > hint * norm(self.c)
         return points, split, n[depth] - n[depth - 1]
 
 
-def algorithm_4 (f, a, b, tol, N_loops=int(1e9)):
+def algorithm_4(f, a, b, tol, N_loops=int(1e9)):
     """ALGORITHM_4 evaluates an integral using adaptive quadrature. The
     algorithm uses Clenshaw-Curtis quadrature rules of increasing
     degree in each interval and bisects the interval if either the
@@ -343,10 +342,12 @@ def algorithm_4 (f, a, b, tol, N_loops=int(1e9)):
             points, split, nr_points_inc = ivals[i_max].refine(f)
             nr_points += nr_points_inc
 
-        if (points[1] - points[0] < points[0] * min_sep
+        if (
+            points[1] - points[0] < points[0] * min_sep
             or points[-1] - points[-2] < points[-2] * min_sep
-            or ivals[i_max].err < (abs(ivals[i_max].igral) * eps
-                                   * Vcond[ivals[i_max].depth])):
+            or ivals[i_max].err
+            < (abs(ivals[i_max].igral) * eps * Vcond[ivals[i_max].depth])
+        ):
             # Remove the interval (while remembering the excess integral and
             # error), since it is either too narrow, or the estimated relative
             # error is already at the limit of numerical accuracy and cannot be
@@ -360,9 +361,12 @@ def algorithm_4 (f, a, b, tol, N_loops=int(1e9)):
             nr_points += nr_points_inc
             if isinstance(result, tuple):
                 raise DivergentIntegralError(
-                    'Possibly divergent integral in the interval'
-                    ' [{}, {}]! (h={})'.format(*result),
-                    ivals[i_max].igral * np.inf, None, nr_points)
+                    "Possibly divergent integral in the interval"
+                    " [{}, {}]! (h={})".format(*result),
+                    ivals[i_max].igral * np.inf,
+                    None,
+                    nr_points,
+                )
             ivals.extend(result)
             ivals[i_max] = ivals.pop()
 
@@ -389,22 +393,25 @@ def algorithm_4 (f, a, b, tol, N_loops=int(1e9)):
             if i_max == len(ivals):
                 i_max = i_min
 
-        if (err == 0
+        if (
+            err == 0
             or err < abs(igral) * tol
-            or (err_excess > abs(igral) * tol
-                and err - err_excess < abs(igral) * tol)
-            or not ivals):
+            or (err_excess > abs(igral) * tol and err - err_excess < abs(igral) * tol)
+            or not ivals
+        ):
             return igral, err, nr_points, ivals
     return igral, err, nr_points, ivals
 
+
 ################ Tests ################
 
+
 def f0(x):
-    return x * np.sin(1/x) * np.sqrt(abs(1 - x))
+    return x * np.sin(1 / x) * np.sqrt(abs(1 - x))
 
 
 def f7(x):
-    return x**-0.5
+    return x ** -0.5
 
 
 def f24(x):
@@ -414,7 +421,7 @@ def f24(x):
 def f21(x):
     y = 0
     for i in range(1, 4):
-        y += 1 / np.cosh(20**i * (x - 2 * i / 10))
+        y += 1 / np.cosh(20 ** i * (x - 2 * i / 10))
     return y
 
 
@@ -439,10 +446,11 @@ def f_one_with_nan(x):
 
 def test_legendre():
     legs = legendre(11)
-    comparisons = [(legs[0], [1], 1),
-                    (legs[1], [0, 1], 1),
-                    (legs[10], [-63, 0, 3465, 0, -30030, 0,
-                                90090, 0, -109395, 0, 46189], 256)]
+    comparisons = [
+        (legs[0], [1], 1),
+        (legs[1], [0, 1], 1),
+        (legs[10], [-63, 0, 3465, 0, -30030, 0, 90090, 0, -109395, 0, 46189], 256),
+    ]
     for a, b, div in comparisons:
         for c, d in zip(a, b):
             assert c * div == d
@@ -450,21 +458,21 @@ def test_legendre():
 
 def test_scalar_product(n=33):
     legs = legendre(n)
-    selection = [0, 5, 7, n-1]
+    selection = [0, 5, 7, n - 1]
     for i in selection:
         for j in selection:
-            assert (scalar_product(legs[i], legs[j])
-                    == ((i == j) and Frac(2, 2*i + 1)))
+            assert scalar_product(legs[i], legs[j]) == ((i == j) and Frac(2, 2 * i + 1))
 
 
 def simple_newton(n):
     """Slower than 'newton()' and prone to numerical error."""
     from itertools import combinations
 
-    nodes = -np.cos(np.arange(n) / (n-1) * np.pi)
-    return [sum(np.prod(-np.asarray(sel))
-                for sel in combinations(nodes, n - d))
-            for d in range(n + 1)]
+    nodes = -np.cos(np.arange(n) / (n - 1) * np.pi)
+    return [
+        sum(np.prod(-np.asarray(sel)) for sel in combinations(nodes, n - d))
+        for d in range(n + 1)
+    ]
 
 
 def test_newton():
@@ -475,8 +483,8 @@ def test_b_def(depth=1):
     legs = [np.array(leg, float) for leg in legendre(n[depth] + 1)]
     result = np.zeros(len(legs[-1]))
     for factor, leg in zip(b_def[depth], legs):
-        factor *= np.sqrt((2*len(leg) - 1) / 2)
-        result[:len(leg)] += factor * leg
+        factor *= np.sqrt((2 * len(leg) - 1) / 2)
+        result[: len(leg)] += factor * leg
     assert_allclose(result, newton(n[depth]), rtol=1e-15)
 
 
@@ -489,11 +497,11 @@ def test_downdate(depth=3):
     fx = np.abs(xi[depth])
     c = _calc_coeffs(fx, depth)
 
-    assert_allclose(c_downdated[:len(c)], c, rtol=0, atol=1e-9)
+    assert_allclose(c_downdated[: len(c)], c, rtol=0, atol=1e-9)
 
 
 def test_integration():
-    old_settings = np.seterr(all='ignore')
+    old_settings = np.seterr(all="ignore")
 
     igral, err, nr_points = algorithm_4(f0, 0, 3, 1e-5)
     assert_allclose(igral, 1.98194117954329, 1e-15)
@@ -537,7 +545,7 @@ def test_analytic(n=200):
     def F(x):
         return F63(x, alpha, beta)
 
-    old_settings = np.seterr(all='ignore')
+    old_settings = np.seterr(all="ignore")
 
     np.random.seed(123)
     params = np.empty((n, 2))
@@ -566,7 +574,7 @@ def test_analytic(n=200):
     np.seterr(**old_settings)
 
 
-if __name__ == '__main__':
+if __name__ == "__main__":
     test_legendre()
     test_scalar_product()
     test_newton()
